from . import utils


class Operators:

    # left = The property of the object being evaluated.
    # right = The value that was entered/selected by the user
    # from the frontend (rule `value` property)

    @staticmethod
    def eval_begins_with(left, right):
        if isinstance(left, list):
            return any(map(lambda x: x.startswith(right), left))
        return False

    @staticmethod
    def eval_between(inputs, bounds):
        if isinstance(inputs, list):
            return any(map(lambda x: bounds[0] < x < bounds[1], inputs))
        return False

    @staticmethod
    def eval_contains(left, right):
        if isinstance(left, list):
            if isinstance(right, list):
                return any([any(map(lambda x: r in x, left)) for r in right])
            return any([right in l for l in left])
        return False

    @staticmethod
    def eval_ends_with(left, right):
        if isinstance(left, list):
            return any(map(lambda x: x.endswith(right), left))
        return False

    @staticmethod
    def eval_equal(left, right):
        if isinstance(left, list):
            return any(map(lambda x: x == right, left))
        return False

    @staticmethod
    def eval_greater(left, right):
        if isinstance(left, list):
            return any(map(lambda x: x > right, left))
        return False

    @staticmethod
    def eval_greater_or_equal(left, right):
        if isinstance(left, list):
            return any(map(lambda x: x >= right, left))
        return False

    @staticmethod
    def eval_in(left, right):
        if isinstance(left, list):
            return all([_ in right for _ in left])
        return False

    @staticmethod
    def eval_is_empty(inputs, _):
        if isinstance(inputs, list):
            return not any(map(lambda x: bool(x), inputs))
        return False

    @staticmethod
    def eval_is_not_empty(inputs, _):
        if isinstance(inputs, list):
            return any(map(lambda x: bool(x), inputs))
        return False

    @staticmethod
    def eval_is_not_null(inputs, _):
        if isinstance(inputs, list):
            return any(map(lambda x: x is not None, inputs))
        return False

    @staticmethod
    def eval_is_null(inputs, _):
        if isinstance(inputs, list):
            return not any(map(lambda x: x is not None, inputs))
        return True

    @staticmethod
    def eval_less(left, right):
        if isinstance(left, list):
            return any(map(lambda x: x < right, left))
        return False

    @staticmethod
    def eval_less_or_equal(left, right):
        if isinstance(left, list):
            return any(map(lambda x: x <= right, left))
        return False

    @staticmethod
    def eval_not_begins_with(left, right):
        if isinstance(left, list):
            return not any(map(lambda x: x.startswith(right), left))
        return False

    @staticmethod
    def eval_not_between(inputs, bounds):
        if isinstance(inputs, list):
            return not any(map(lambda x: bounds[0] < x < bounds[1], inputs))
        return False

    @staticmethod
    def eval_not_contains(left, right):
        if isinstance(left, list):
            if isinstance(right, list):
                return not any(
                    [any(map(lambda x: r in x, left)) for r in right]
                )
            return not any([right in l for l in left])
        return False

    @staticmethod
    def eval_not_ends_with(left, right):
        if isinstance(left, list):
            return not any(map(lambda x: x.endswith(right), left))
        return False

    @staticmethod
    def eval_not_equal(left, right):
        if isinstance(left, list):
            return not any(map(lambda x: x == right, left))
        return False

    @staticmethod
    def eval_not_in(left, right):
        if isinstance(left, list):
            return not all([_ in right for _ in left])
        return False
    
    @staticmethod
    def eval_length_equal(left, right):
        if isinstance(left, list):
            length = len(right) if isinstance(right, list) else int(right)
            return len(left) == length
        return False

    @staticmethod
    def eval_length_not_equal(left, right):
        if isinstance(left, list):
            length = len(right) if isinstance(right, list) else int(right)
            return len(left) != length
        return False
    
    @staticmethod
    def eval_length_greater(left, right):
        if isinstance(left, list):
            length = len(right) if isinstance(right, list) else int(right)
            return len(left) > length
        return False

    @staticmethod
    def eval_length_greater_or_equal(left, right):
        if isinstance(left, list):
            length = len(right) if isinstance(right, list) else int(right)
            return len(left) >= length
        return False

    @staticmethod
    def eval_length_less(left, right):
        if isinstance(left, list):
            length = len(right) if isinstance(right, list) else int(right)
            return len(left) < length
        return False

    @staticmethod
    def eval_length_less_or_equal(left, right):
        if isinstance(left, list):
            length = len(right) if isinstance(right, list) else int(right)
            return len(left) <= length
        return False

    @staticmethod
    def eval_occurrence(left, right):
        """ 

        This funtion compares the number of occurrences of a value with an
        integer value and returns a boolean indicating whether the condition
        is true or false

        Args:
            left (list): list of values extracted from the inputs object.
            right (str): string that represents the condition that we want to 
                check about the occurences. It must match this pattern: 
                <value> <operator> <int>
        
        Returns:
            bool: a boolean value whether the condition is true or false
        """
        if isinstance(left, list):
            if utils.validate_string(right):
                left_operand, operator, right_operand = (
                    utils.split_string_by_operator(right)
                )
                occurence = left.count(left_operand)
                return operator(occurence, right_operand)
            return False
        return False


from datetime import datetime

from pytimeparse.timeparse import timeparse

from jqqb.operators import Operators


class Rule:
    def __init__(self, rule_dict):
        self.id = rule_dict["id"]
        self.field = rule_dict["field"]
        self.type = rule_dict["type"]
        self.input = rule_dict["input"]
        self.operator = rule_dict["operator"]
        self.value = rule_dict["value"]
        self.value_type = rule_dict["value_type"]

    def evaluate(self, obj):
        results = []
        result = self.get_operator()(
            self.get_input(obj, results),
            self.get_value()
        )
        return result

    def inputs(self, obj):
        results = []
        inputs = self.get_input(obj, results)
        return inputs

    def values(self):
        values = self.get_value()
        return values

    def inspect(self, obj):
        return self.inputs(obj), self.values(), self.evaluate(obj)

    def get_operator(self):
        return getattr(Operators, "eval_" + self.operator)

    def get_input(self, obj, results):
        fields = self.field.split(".")
        fd_index = 0

        if isinstance(obj, list):
            for i in range(len(obj)):
                self.get_input(obj[i], results)

        elif isinstance(obj, dict):
            while (fd_index < len(fields) and fields[fd_index] not in obj):
                fd_index += 1

            if fd_index < len(fields) and fields[fd_index] in obj:
                self.get_input(obj[fields[fd_index]], results)

        else:
            results.append(self.typecast_value(obj, type=self.type))

        results = [x for x in results if x is not None] or None
        return results

    def get_value(self):
        if isinstance(self.value, list):
            return list(
                map(
                    lambda x: self.typecast_value(x, type=self.value_type), 
                    self.value
                )
            )
        return self.typecast_value(self.value, type=self.value_type)

    def typecast_value(self, value_to_cast, type):
        if value_to_cast is None:
            return None

        if type == "string":
            return str(value_to_cast)
        elif type == "integer":
            return int(value_to_cast)
        elif type == "double":
            return float(value_to_cast)
        elif type == "datetime":
            return (
                datetime.fromisoformat(value_to_cast)
                if isinstance(value_to_cast, str)
                else value_to_cast
            )
        elif type == "date":
            return (
                datetime.strptime(value_to_cast, "%Y-%m-%d")
                if isinstance(value_to_cast, str)
                else value_to_cast
            )
        elif type == "time":
            return (
                timeparse(value_to_cast)
                if isinstance(value_to_cast, str)
                else value_to_cast
            )
        elif type == "boolean":
            if isinstance(value_to_cast, str):
                return value_to_cast.lower() in [
                    "true",
                    "1",
                    "t",
                    "y",
                    "yes",
                    "yeah",
                    "yup",
                    "certainly",
                    "uh-huh",
                ]
        elif type == "list":
            ...
        return value_to_cast
